"""API Views and ViewSets"""

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models.loading import get_models
from django.http import Http404
from django.template.defaultfilters import slugify
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from firebase_token_generator import create_token
import elasticsearch
from rest_framework import (
    decorators,
    filters,
    status,
    viewsets,
    routers
)
from rest_framework.decorators import detail_route
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response
from elastimorphic.models import polymorphic_indexable_registry
from elasticutils.contrib.django import get_es

from bulbs.content.models import Content, Tag, LogEntry, FeatureType, ObfuscatedUrlInfo
from bulbs.content.serializers import (
    LogEntrySerializer, PolymorphicContentSerializer,
    TagSerializer, UserSerializer, FeatureTypeSerializer,
    ObfuscatedUrlInfoSerializer
)
from bulbs.contributions.serializers import ContributionSerializer
from bulbs.contributions.models import Contribution
from .mixins import UncachedResponse
from .permissions import CanEditContent, CanPublishContent


class ContentViewSet(UncachedResponse, viewsets.ModelViewSet):
    """
    uncached viewset for the `bulbs.content.Content` model
    """

    model = Content
    queryset = Content.objects.select_related("tags").all()
    serializer_class = PolymorphicContentSerializer
    include_base_doctype = False
    paginate_by = 20
    filter_fields = ("tags", "authors", "feature_types", "published", "types")
    search_fields = ("title", "description")
    permission_classes = [IsAdminUser, CanEditContent]

    def get_serializer_class(self):
        """gets the class type of the serializer

        :return: `rest_framework.Serializer`
        """
        klass = None

        if getattr(self, "object", None):
            klass = self.object.__class__
        elif "doctype" in self.request.REQUEST:
            try:
                klass = Content.get_doctypes()[self.request.REQUEST["doctype"]]
            except KeyError:
                raise Http404

        if hasattr(klass, "get_serializer_class"):
            return klass.get_serializer_class()

        # TODO: fix deprecation warning here -- `get_serializer_class` is going away soon!
        return super(ContentViewSet, self).get_serializer_class()

    def post_save(self, obj, created=False):
        """indexes the object to ElasticSearch after any save function (POST/PUT)

        :param obj: instance of the saved object
        :param created: boolean expressing if object is newly created (`False` if updated)
        :return: `rest_framework.viewset.ModelViewSet.post_save`
        """
        from bulbs.content.tasks import index

        index.delay(obj.polymorphic_ctype_id, obj.pk)

        message = "Created" if created else "Saved"
        LogEntry.objects.log(self.request.user, obj, message)
        return super(ContentViewSet, self).post_save(obj, created=created)

    def list(self, request, *args, **kwargs):
        """I'm overriding this so that the listing pages can be driven from ElasticSearch

        :param request: a WSGI request object
        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """

        # check for a text query
        search_kwargs = {"published": False}
        search_query = request.QUERY_PARAMS.get("search", None)
        if search_query:
            search_kwargs["query"] = search_query

        if "before" in request.QUERY_PARAMS:
            search_kwargs["before"] = parse_datetime(request.QUERY_PARAMS["before"])

        if "after" in request.QUERY_PARAMS:
            search_kwargs["after"] = parse_datetime(request.QUERY_PARAMS["after"])

        if "status" in request.QUERY_PARAMS:
            search_kwargs["status"] = request.QUERY_PARAMS["status"]
            del search_kwargs["published"]

        # filter on specific fields using `filter_fields` on your view
        filter_fields = getattr(self, "filter_fields", None)
        if filter_fields:
            for field_name in filter_fields:
                filter_query = request.QUERY_PARAMS.getlist(field_name, None)
                if filter_query:

                    # We need to figure out how to match on a slug, not a name
                    if field_name == "feature_types":
                        filter_query = [slugify(f) for f in filter_query]

                    search_kwargs[field_name] = filter_query

        self.object_list = self.model.search_objects.search(**search_kwargs).full()

        # Switch between paginated or standard style responses
        page = self.paginate_queryset(self.object_list)
        if page is not None:
            serializer = self.get_pagination_serializer(page)
        else:
            serializer = self.get_serializer(self.object_list, many=True)

        return Response(serializer.data)

    @decorators.action(permission_classes=[CanPublishContent])
    def publish(self, request, **kwargs):
        """sets the `published` value of the `Content`

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        content = self.get_object()

        if "published" in request.DATA:
            if not request.DATA["published"]:
                content.published = None
            else:
                publish_dt = parse_datetime(request.DATA["published"])
                if publish_dt:
                    publish_dt = publish_dt.astimezone(timezone.utc)
                else:
                    publish_dt = None
                content.published = publish_dt
        else:
            content.published = timezone.now()

        content.save()
        LogEntry.objects.log(request.user, content, content.get_status())
        return Response({"status": content.get_status(), "published": content.published})

    @decorators.action(permission_classes=[CanPublishContent])
    def trash(self, request, **kwargs):
        """destroys a `Content` instance and removes it from the ElasticSearch index

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        :raise: 404
        """
        content = self.get_object()

        content.indexed = False
        content.save()

        es = get_es(urls=settings.ES_URLS)
        try:
            es.delete(index=content.get_index_name(), doc_type=content.get_mapping_type_name(),
                      id=content.id)
            LogEntry.objects.log(request.user, content, "Trashed")
            return Response({"status": "Trashed"})
        except elasticsearch.exceptions.NotFoundError:
            return Response(status=status.HTTP_404_NOT_FOUND)

    @decorators.link()
    def status(self, request, **kwargs):
        """This endpoint returns a status text, currently one of:
          - "Draft" (If no publish date is set, and no item exists in the editor queue)
          - "Waiting for Editor" (If no publish date is set, and an item exists in the editor queue)
          - "Published" (The published date is in the past)
          - "Scheduled" (The published date is set in the future)

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        content = self.get_object()
        return Response({"status": content.get_status()})

    @detail_route(methods=["post", "get"])
    def contributions(self, request, **kwargs):
        """gets or adds contributions

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        # Check if the contribution app is installed
        if Contribution not in get_models():
            return Response([])
        queryset = Contribution.objects.filter(content=self.get_object())
        if request.method == "POST":
            serializer = ContributionSerializer(
                queryset,
                data=request.DATA,
                many=True,
                allow_add_remove=True)
            if not serializer.is_valid():
                return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
            serializer.save()
            return Response(serializer.data)
        else:
            serializer = ContributionSerializer(queryset, many=True)
            return Response(serializer.data)

    @detail_route(methods=["post"], permission_classes=[CanEditContent])
    def create_token(self, request, **kwargs):
        """Create a new obfuscated url info to use for accessing unpublished content.

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """

        data = ObfuscatedUrlInfoSerializer(ObfuscatedUrlInfo.objects.create(
            content=self.get_object(),
            create_date=request.DATA["create_date"],
            expire_date=request.DATA["expire_date"]
        )).data

        return Response(data, status=status.HTTP_200_OK, content_type="application/json")

    @detail_route(methods=["get"], permission_classes=[CanEditContent])
    def list_tokens(self, request, **kwargs):
        """List all tokens for this content instance.

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """

        # no date checking is done here to make it more obvious if there's an issue with the
        # number of records. Date filtering will be done on the frontend.
        infos = [ObfuscatedUrlInfoSerializer(info).data
                 for info in ObfuscatedUrlInfo.objects.filter(content=self.get_object())]
        return Response(infos, status=status.HTTP_200_OK, content_type="application/json")


class TagViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached viewset for the `bulbs.content.Tag` model
    """

    model = Tag
    queryset = Tag.objects.all()
    serializer_class = TagSerializer
    search_fields = ("name",)
    paginate_by = 50

    def list(self, request, *args, **kwargs):
        """I'm overriding this so that the listing pages can be driven from ElasticSearch

        :param request: a WSGI request object
        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """

        search_query = Tag.search_objects.s()
        if "search" in request.REQUEST:
            search_query = search_query.query(
                name__match_phrase=request.REQUEST["search"], should=True
            ).query(
                name__term=request.REQUEST["search"], should=True
            )

        if "types" in request.REQUEST:
            search_query = search_query.doctypes(*request.REQUEST.getlist("types"))

        # HACK ALERT. I changed the edge ngram to go from 3 to 10, so "TV" got screwed
        if len(request.REQUEST.get("search", [])) < 3:
            self.object_list = Tag.objects.filter(
                name__istartswith=request.REQUEST["search"].lower()
            )
            if "types" in request.REQUEST:
                type_classes = [polymorphic_indexable_registry.all_models[type_name] for type_name
                                in request.REQUEST.getlist("types")]
                print(type_classes)
                self.object_list = self.object_list.instance_of(*type_classes)
        else:
            self.object_list = search_query.full()

        # Switch between paginated or standard style responses
        page = self.paginate_queryset(self.object_list)
        if page is not None:
            serializer = self.get_pagination_serializer(page)
        else:
            serializer = self.get_serializer(self.object_list, many=True)

        return Response(serializer.data)


class UserViewSet(UncachedResponse, viewsets.ModelViewSet):
    """
    uncached viewset for the User model -- User can be whatever is defined as the default system
    user model (`auth.User` or any other custom model)
    """

    model = get_user_model()
    serializer_class = UserSerializer
    filter_backends = (filters.SearchFilter, filters.OrderingFilter)
    search_fields = ("first_name", "last_name", "username")
    paginate_by = 20


class LogEntryViewSet(UncachedResponse, viewsets.ModelViewSet):
    """
    uncached viewset for `bulbs.content.LogEntry` model
    """

    model = LogEntry
    serializer_class = LogEntrySerializer

    def get_queryset(self):
        """creates the base queryset object for the serializer

        :return: an instance of `django.db.models.QuerySet`
        """
        qs = super(LogEntryViewSet, self).get_queryset()
        content_id = self.request.QUERY_PARAMS.get("content", None)
        if content_id:
            qs = qs.filter(object_id=content_id)
        return qs

    def create(self, request, *args, **kwargs):
        """
        Grabbing the user from request.user, and the rest of the method
        is the same as ModelViewSet.create().

        :param request: a WSGI request object
        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        :raise: 400
        """
        data = request.DATA.copy()
        data["user"] = request.user.id
        serializer = self.get_serializer(data=data, files=request.FILES)

        if serializer.is_valid():
            self.pre_save(serializer.object)
            self.object = serializer.save(force_insert=True)
            self.post_save(self.object, created=True)
            headers = self.get_success_headers(serializer.data)
            return Response(serializer.data, status=status.HTTP_201_CREATED,
                            headers=headers)

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class AuthorViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached readonly viewset for the system user model limited to author groups and permissions
    """

    model = get_user_model()
    serializer_class = UserSerializer
    filter_backends = (filters.SearchFilter, filters.OrderingFilter)
    search_fields = ("first_name", "last_name", "username")

    def get_queryset(self):
        """created the base queryset object for the serializer limited to users within the authors
        groups and having `is_staff`

        :return: `django.db.models.QuerySet`
        """
        author_filter = getattr(settings, "BULBS_AUTHOR_FILTER", {"is_staff": True})
        queryset = self.model.objects.filter(**author_filter).distinct()
        return queryset


class FeatureTypeViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached readonly viewset for the `bulbs.content.FeatureType` model
    """

    model = FeatureType
    serializer_class = FeatureTypeSerializer
    filter_backends = (filters.SearchFilter, filters.OrderingFilter)
    search_fields = ("name", )


class MeViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached readonly viewset for users to get information about themselves
    """

    def retrieve(self, request, *args, **kwargs):
        """gets basic information about the user

        :param request: a WSGI request object
        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        data = UserSerializer().to_native(request.user)

        # add superuser flag only if user is a superuser, putting it here so users can only
        # tell if they are themselves superusers
        if request.user.is_superuser:
            data['is_superuser'] = True

        # attempt to add a firebase token if we have a firebase secret
        secret = getattr(settings, 'FIREBASE_SECRET', None)
        if secret:
            # use firebase auth to provide auth variables to firebase security api
            firebase_auth_payload = {
                'id': request.user.pk,
                'username': request.user.username,
                'email': request.user.email,
                'is_staff': request.user.is_staff
            }
            data['firebase_token'] = create_token(secret, firebase_auth_payload)

        return Response(data)


# api router for aforementioned/defined viewsets
# note: me view is registered in urls.py
api_v1_router = routers.DefaultRouter()
api_v1_router.register(r"content", ContentViewSet, base_name="content")
api_v1_router.register(r"tag", TagViewSet, base_name="tag")
api_v1_router.register(r"log", LogEntryViewSet, base_name="logentry")
api_v1_router.register(r"author", AuthorViewSet, base_name="author")
api_v1_router.register(r"feature-type", FeatureTypeViewSet, base_name="feature-type")
api_v1_router.register(r"user", UserViewSet, base_name="user")



from elastimorphic.serializers import ContentTypeField

from rest_framework import serializers

from bulbs.content.models import Content
from bulbs.content.serializers import ContentSerializer

from .models import PZone
from .operations import PZoneOperation, InsertOperation, DeleteOperation, ReplaceOperation


class PZoneField(serializers.WritableField):
    def field_to_native(self, obj, field_name):
        data = []
        for content in obj:
            data.append(ContentSerializer(instance=content).data)

        return data

    def from_native(self, data):
        return [{"id": content_data["id"]} for content_data in data]


class PZoneSerializer(serializers.ModelSerializer):
    class Meta:
        model = PZone
        exclude = ("data",)

    content = PZoneField(source="data")


class _PZoneOperationSerializer(serializers.ModelSerializer):
    """Parent class of pzone operation serializers."""

    type_name = ContentTypeField(source="polymorphic_ctype_id")
    pzone = serializers.PrimaryKeyRelatedField()
    when = serializers.DateTimeField()
    applied = serializers.BooleanField(default=False)
    content = serializers.PrimaryKeyRelatedField()
    content_title = serializers.SerializerMethodField('get_content_title')

    class Meta:
        model = PZoneOperation

    def get_content_title(self, obj):
        """Get content's title."""
        return Content.objects.get(id=obj.content.id).title


class InsertOperationSerializer(_PZoneOperationSerializer):

    index = serializers.IntegerField(min_value=0)

    class Meta:
        model = InsertOperation


class ReplaceOperationSerializer(_PZoneOperationSerializer):

    index = serializers.IntegerField(min_value=0)

    class Meta:
        model = ReplaceOperation


class DeleteOperationSerializer(_PZoneOperationSerializer):

    class Meta:
        model = DeleteOperation



import json

from django.contrib.auth import get_user_model
# from django.conf import settings
# from django.db.models.loading import get_model
from django.test import TestCase
from django.test.client import Client


class VideoAPICase(TestCase):

    def setUp(self):
        User = get_user_model()
        # User = get_model(*settings.AUTH_USER_MODEL.split("."))
        admin = User.objects.create_user("admin", "tech@theonion.com", "secret")
        admin.is_staff = True
        admin.save()

    def test_video_create(self):

        client = Client()
        client.login(username='admin', password='secret')

        data = {
            "name": "Testing Video"
        }
        response = client.post(
            "/videos/api/video/",
            json.dumps(data), content_type="application/json")
        self.assertEqual(response.status_code, 201)  # 201 Created
        self.assertEqual(response.data.get("name"), "Testing Video")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import json_field.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Video',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('data', json_field.fields.JSONField(default='null', help_text=b'This is JSON that is returned from an encoding job', null=True, blank=True)),
                ('sources', json_field.fields.JSONField(default=[], help_text=b'This is a JSON array of sources.', null=True, blank=True)),
                ('poster_url', models.URLField(null=True, blank=True)),
                ('status', models.IntegerField(default=0, choices=[(0, b'Not started'), (1, b'Complete'), (2, b'In Progress'), (3, b'Failed')])),
                ('job_id', models.IntegerField(help_text=b'The zencoder job ID', null=True, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



"""
This module stores code for Shallow content results.

In the future, we should remove this, and pass
simple dictionaries into our templates.
"""

from datetime import datetime

from elasticutils import SearchResults
from django.utils import timezone
from django.utils.timezone import utc
import bulbs.content

from elastimorphic.base import PolymorphicS, ModelSearchResults


class ShallowFeatureType(object):

    def __init__(self, name, slug=None):
        self.name = name
        self.slug = slug

    def __str__(self):
        return self.name

    def __unicode__(self):
        return self.name


class ShallowTag(object):

    def __init__(self, data):
        self.id   = data.get('id')
        self.name = data.get('name')
        self.slug = data.get('slug')


class ShallowAuthor(object):

    def __init__(self, data):
        self.id = data.get('id')
        self.username = data.get('username')
        self.first_name = data.get('first_name')
        self.last_name = data.get('last_name')


class ShallowAuthorRelation(list):

    def __init__(self, items):
        for item in items:
            self.append(ShallowAuthor(item))

    def all(self):
        return self


class ShallowTagRelation(list):

    def __init__(self, items):
        for item in items:
            self.append(ShallowTag(item))

    def all(self):
        return self


class ShallowObject(object):

    def __init__(self, data):
        for key, item in data.items():
            if hasattr(self, key):
                continue
            if isinstance(item, dict):
                setattr(self, key, ShallowObject(item))
            elif isinstance(item, list):
                setattr(self, key, ShallowRelation(item))
            else:
                setattr(self, key, item)


class ShallowRelation(list):
    def __init__(self, items):
        for item in items:
            if isinstance(item, dict):
                self.append(ShallowObject(item))
            else:
                self.append(item)

    def all(self):
        return self


class ShallowContentResult(ShallowObject):

    def __init__(self, _source, type=None):
        if type:
            self.type = type

        published_utc = None
        if _source.get('published'):
            try:
                published_utc = datetime.strptime(
                    _source.get('published'), '%Y-%m-%dT%H:%M:%S.%f+00:00'
                ).replace(tzinfo=utc)
            except ValueError:
                try:
                    published_utc = datetime.strptime(
                        _source.get('published'), '%Y-%m-%dT%H:%M:%S+00:00'
                    ).replace(tzinfo=utc)
                except ValueError:
                    pass

        if published_utc:
            self.published = timezone.localtime(published_utc)
        else:
            self.published = None

        self.feature_type = ShallowFeatureType(
            _source.get('feature_type').get('name'),
            slug=_source.get('feature_type').get('slug'))
        super(ShallowContentResult, self).__init__(_source)

    def __unicode__(self):
        return "<Content>"

    def __repr__(self):
        return "<Content>"

    def get_absolute_url(self):
        return getattr(self, 'absolute_url', None)

    @property
    def is_published(self):
        if self.published:
            now = timezone.now()
            if now >= self.published:
                return True
        return False

    def ordered_tags(self):
        tags = list(self.tags.all())
        sorted(tags, key=lambda tag: ((tag.type != "content_tag") * 100000) + bulbs.content.TagCache.count(tag.slug))
        return tags


class ShallowContentSearchResults(SearchResults):

    def set_objects(self, results):
        self.objects = []
        for r in results:
            self.objects.append(ShallowContentResult(r['_source'], type=r['_type']))

    def __iter__(self):
        return self.objects.__iter__()


class ShallowContentS(PolymorphicS):

    def get_results_class(self):
        if self.as_models:
            return ModelSearchResults
        return ShallowContentSearchResults



import json
import base64
import hmac
import hashlib
import datetime

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpResponseNotAllowed
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.cache import cache_control


from .models import Video


@cache_control(no_cache=True)
@csrf_exempt
def notification(request):
    if request.method != "POST":
        return HttpResponseNotAllowed(["GET"])

    # Grab the encoding details from message
    details = json.loads(request.body)

    sources = []
    """We need to munge the encoding details into an easily rendered list of sources:
    {
        "url"         : "http://example.com/some_video.mp4",
        "content_type": "video/mp4",
        "height"      : 320,
        "width"       : 640,
        "bitrate"     : 120
    }

    "url" and "content_type" are the only required properties, with "width", "height" and "bitrate"
    being only available for some items.
    """
    for output in details.get('outputs', []):
        for filename in settings.VIDEO_ENCODING.get('sources', {}).keys():
            if output.get('url').endswith(filename):
                source = {
                    'url': output['url'],
                    'content_type': settings.VIDEO_ENCODING.get('sources', {}).get(filename)
                }
                if output.get('height'):
                    source['height'] = output['height']

                if output.get('width'):
                    source['width'] = output['width']

                if output.get('video_bitrate_in_kbps'):
                    source['bitrate'] = output['video_bitrate_in_kbps']

                sources.append(source)
                break

    job_id = details['job']['id']
    videos = Video.objects.filter(job_id=job_id)
    if videos.count() == 1:
        video = videos[0]
        video.data = details
        video.sources = sources

        if details['job']['state'] == 'finished':
            video.status = Video.COMPLETE
        if details['job']['state'] in ['cancelled', 'failed']:
            video.status = Video.FAILED
        video.save()
    return HttpResponse(status=204)


@cache_control(no_cache=True)
@staff_member_required
def video_attrs(request):
    expiration = datetime.datetime.utcnow() + datetime.timedelta(hours=1)
    directory = settings.VIDEO_ENCODING.get('directory')
    # callback_url = 'http://' + request.get_host() + reverse('videoads.videos.views.upload_successful')

    policy_dict = {
        "expiration": expiration.strftime('%Y-%m-%dT%H:%M:%S.000Z'),
        "conditions": [
            {"bucket": settings.VIDEO_ENCODING.get('bucket')},
            {"acl": "private"},
            {"success_action_status": '201'},
            ["starts-with", "$key", directory],
            ["content-length-range", 0, 1073741824],
        ]
    }

    policy_document = json.dumps(policy_dict)
    policy = base64.b64encode(policy_document)
    signature = base64.b64encode(
        hmac.new(settings.AWS_SECRET_ACCESS_KEY, policy, hashlib.sha1).digest())

    contents = {
        'AWSAccessKeyId': settings.AWS_ACCESS_KEY_ID,
        'acl': 'private',
        'success_action_status': '201',
        'policy': policy,
        'signature': signature,
        'Content-Type': '$Content-Type',
        'bucket': settings.VIDEO_ENCODING.get('bucket'),
        'directory': directory,
        'zencoderApiKey': settings.VIDEO_ENCODING.get('zencoder_api_key')
    }
    response_string = "var videoAttrs = %s;" % json.dumps(contents)
    return HttpResponse(response_string, content_type="application/javascript")
